// (C) 2003 Ulrich Hecht <uli@emulinks.de>
// if2a patch by D. Gauchard <deyv@free.fr>
// Licensed under the terms of the GNU Public License version 2

#include <string.h>
#include <stdio.h>
#include <usb.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>

#include  <linux/version.h>
#include  <linux/usb.h>
#include  <linux/usbdevice_fs.h>
#include  <sys/ioctl.h>

#include "f2a.h"

#include "f2afirm.hex-src.h"

//#define EZ "/proc/ezusb/dev0"

char* EZ;

struct usb_bus *busses;
struct usb_bus *bus;
struct usb_device *dev;
struct usb_device *f2a=NULL;
usb_dev_handle *f2ahandle;

#define RW_INTERNAL     0xA0            /* hardware implements this one */
#define RW_EEPROM       0xA2
#define RW_MEMORY       0xA3
#define GET_EEPROM_SIZE 0xA5

typedef enum {
	_undef = 0,
	internal_only,              /* hardware first-stage loader */
	skip_internal,              /* first phase, second-stage loader */
	skip_external               /* second phase, second-stage loader */
} ram_mode;


struct ram_poke_context {
	int         device;
	ram_mode    mode;
	unsigned    total, count;
};  

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,3)
/*
 * in 2.5, "struct usbdevfs_ctrltransfer" fields were renamed
 * to match the USB spec
*/
#       define bRequestType     requesttype
#       define bRequest         request
#       define wValue           value
#       define wIndex           index
#       define wLength          length
#endif


static inline int ctrl_msg (
    int                                 device,
    unsigned char                       requestType,
    unsigned char                       request,
    unsigned short                      value,
    unsigned short                      index,
    unsigned char                       *data,
    size_t                              length
) {
    struct usbdevfs_ctrltransfer        ctrl;

    if (length > USHRT_MAX) {
        fputs ("length too big", stderr);
        return -EINVAL;
    }

    /* 8 bytes SETUP */
    ctrl.bRequestType = requestType;
    ctrl.bRequest = request;
    ctrl.wValue   = value;
    ctrl.wLength  = (unsigned short) length; 
    ctrl.wIndex = index;

    /* "length" bytes DATA */
    ctrl.data = data; 

    ctrl.timeout = 10000;

    return ioctl (device, USBDEVFS_CONTROL, &ctrl); 
}


/*
 * Modifies the CPUCS register to stop or reset the CPU.
 * Returns false on error.
 */
static int ezusb_cpucs (
    int                 device,
    unsigned short      addr,   
    int                 doRun
) {
    int                 status;
    unsigned char       data = doRun ? 0 : 1;

    if (verbose)
        fprintf (stderr, "%s\n", data ? "stop CPU" : "reset CPU");
    status = ctrl_msg (device,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        RW_INTERNAL,
        addr, 0,
        &data, 1);
    if (status != 1) { 
        char *mesg = "can't modify CPUCS";
        if (status < 0)
            perror (mesg); 
        else
            fprintf (stderr, "%s\n", mesg);
        return 0;
    } else  
        return 1;
}



/*
 * Issues the specified vendor-specific write request.
 */
static int ezusb_write (
    int                                 device,
    char                                *label,
    unsigned char                       opcode,
    unsigned short                      addr,
    const unsigned char                 *data,
    size_t                              len
) {
    int                                 status;

    if (verbose)
        fprintf (stderr, "%s, addr 0x%04x len %4d (0x%04x)\n",
                label, addr, len, len);
    status = ctrl_msg (device,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE, opcode, 
        addr, 0,
        (unsigned char *) data, len);
    if (status != len) {
        if (status < 0)
            perror (label);
        else
            fprintf (stderr, "%s ==> %d\n", label, status);
    }
    return status; 
}



#define RETRY_LIMIT 5

static int poke (
		void                *context,
		unsigned short      addr,
		int                 external,
		const unsigned char *data,
		size_t              len
		) {
	struct ram_poke_context     *ctx = context;
	int                 rc;
	unsigned            retry = 0;

	switch (ctx->mode) {
		case internal_only:         /* CPU should be stopped */
			if (external) {
				fprintf (stderr, "can't write %d bytes external memory at 0x%04x\n", len, addr);
				return -EINVAL;
			}
			break;
		case skip_internal:         /* CPU must be running */
			if (!external) {
				if (verbose >= 2) 
					fprintf (stderr, "SKIP on-chip RAM, %d bytes at 0x%04x\n", len, addr);
				return 0;
			}
			break;
		case skip_external:         /* CPU should be stopped */
			if (external) {
				if (verbose >= 2) 
					fprintf (stderr, "SKIP external RAM, %d bytes at 0x%04x\n", len, addr);
				return 0;
			}
			break;
		default:
			fprintf (stderr, "bug\n");
			return -EDOM;
	}

	ctx->total += len;
	ctx->count++;

	/* Retry this till we get a real error. Control messages are not
	 * NAKed (just dropped) so time out means is a real problem.
	 */
	while ((rc = ezusb_write (ctx->device,
					external ? "write external" : "write on-chip",
					external ? RW_MEMORY : RW_INTERNAL,
					addr, data, len)) < 0
			&& retry < RETRY_LIMIT) {
		if (errno != ETIMEDOUT)
			break;
		retry += 1;
	}
	return (rc < 0) ? -errno : 0;
}



int is_external (unsigned short addr, size_t len)
{
	/* with 8KB RAM, 0x0000-0x1b3f can be written
	 * we can't tell if it's a 4KB device here
	 */
	if (addr <= 0x1b3f)
		return ((addr + len) > 0x1b40);

	/* there may be more RAM; unclear if we can write it.
	 * some bulk buffers may be unused, 0x1b3f-0x1f3f
	 * firmware can set ISODISAB for 2KB at 0x2000-0x27ff
	 */
	return 1;
}



unsigned char* skip_line(unsigned char* buffer) {
	while ((*buffer != '\n' || *buffer != '\r') && *buffer != '\0') 
		buffer++;
	if (*buffer == '\r' || *buffer == '\n' ) buffer++;
	return buffer;
}


int parse_ihex (
		unsigned char *image,
		void        *context
		) {

	unsigned char       data [1023];
	unsigned short      data_addr = 0;
	size_t              data_len = 0;
	int                 rc = -1;
	int                 first_line = 1;
	int                 external = 0;

	char            *buf, *cp;

	buf = image;
	for (;;) {
		char            tmp, type;
		size_t          len;
		unsigned        idx, off;


		/* test if end of buffer reached */
		if (buf[0] == '\0') {
			fprintf (stderr, "not an ihex record: %s", buf);
			return -2;
		}

		/* skip cariage return */
		while (buf[0] == '\r' || buf[0] == '\n') buf++;

		/* EXTENSION: "# comment-till-end-of-line", for copyrights etc */
		if (buf[0] == '#') {
			buf = skip_line(buf);
			continue;
		}

		if (buf[0] != ':') {
			fprintf (stderr, "not an ihex record: %s\n", buf);

			fprintf(stderr, "{%x}\n", *buf);

			return -2;
		}

		/* Read the length field (up to 16 bytes) */
		tmp = buf[3];
		buf[3] = 0;
		len = strtoul(buf+1, 0, 16);
		buf[3] = tmp;

		/* Read the target offset (address up to 64KB) */
		tmp = buf[7];
		buf[7] = 0;
		off = strtoul(buf+3, 0, 16);
		buf[7] = tmp;

		/* Initialize data_addr */
		if (first_line) {
			data_addr = off;
			first_line = 0;
		}

		/* Read the record type */
		tmp = buf[9];
		buf[9] = 0;
		type = strtoul(buf+7, 0, 16);
		buf[9] = tmp;

		/* fprintf(stderr, "%d - %d - %d\n", off, len, type); */

		/* If this is an EOF record, then make it so. */
		if (type == 1) {
			if (verbose > 2)
				fprintf(stderr, "Read EOF entry\n");
			break;
		}

		if (type != 0) {
			fprintf (stderr, "unsupported record type: %u\n", type);
			return -3;
		}

		/* flush the saved data if it's not contiguous,
		 * or when we've buffered as much as we can.
		 */
		if (data_len != 0
				&& (off != (data_addr + data_len)
					// || !merge
					|| (data_len + len) > sizeof data)) {
			external = is_external (data_addr, data_len);
			rc = poke (context, data_addr, external, data, data_len);

			if (rc < 0)
				return -1;
			if (verbose > 2)
				fprintf(stderr, "Write %d bytes to ezusb\n", data_len); 
			data_addr = off;
			data_len = 0;
		}

		/* append to saved data, flush later */
		for (idx = 0, cp = buf+9 ;  idx < len ;  idx += 1, cp += 2) {
			tmp = cp[2];
			cp[2] = 0;
			data [data_len + idx] = strtoul(cp, 0, 16);
			cp[2] = tmp;
		}
		buf += len*2 + 11;
		data_len += len;
	}


	/* flush any data remaining */
	if (data_len != 0) {
		external = is_external (data_addr, data_len);
		rc = poke (context, data_addr, external, data, data_len);
		if (rc < 0)
			return -1;
		if (verbose > 2)
			fprintf(stderr, "Write %d bytes to ezusb\n", data_len);
	}
	return 0;
}




int ezusb_load_ram (int fd)
{
	unsigned short              cpucs_addr;
	struct ram_poke_context     ctx;
	int                         status;

	cpucs_addr = 0x7f92;

	ctx.mode = internal_only;

	/* don't let CPU run while we overwrite its code/data */
	if (!ezusb_cpucs (fd, cpucs_addr, 0))
		return -1;

	/* scan the image, first (maybe only) time */
	ctx.device = fd;
	ctx.total = ctx.count = 0;
	status = parse_ihex (f2afirm_hex, &ctx);
	if (status < 0) {
		fprintf (stderr, "unable to download firmware");
		return status;
	}

	if (verbose)
		fprintf (stderr, "... WROTE: %d bytes, %d segments, avg %d\n",
				ctx.total, ctx.count, ctx.total / ctx.count);

	/* now reset the CPU so it runs what we just downloaded */
	if (!ezusb_cpucs (fd, cpucs_addr, 1))
		return -1;

	return 0;
}

int usb_connect()
{
	int result;
	int firmware_loaded=0;
	int status;
	int fd;

	EZ = (char*) malloc(sizeof(char)*30);
	memset(EZ, 0, 30);
	
	usb_init();
	usb_find_busses();

find_f2a:
	usb_find_devices();
	busses=usb_busses;

	for (bus=busses; bus; bus=bus->next) {
		for (dev=bus->devices; dev; dev=dev->next) {
			if ((dev->descriptor.idVendor == 0x547) &&
					(dev->descriptor.idProduct == 0x1002)) {
				f2a=dev;
				printf("Ezusb Firmware is loaded.\n");
				snprintf(EZ, 22, "/proc/bus/usb/%s/%s", bus->dirname, dev->filename);
			}
			if ((dev->descriptor.idVendor == 0x547) &&
					(dev->descriptor.idProduct == 0x2131) &&
					!firmware_loaded)
			{
				printf("loading Ezusb Firmware... ");
				snprintf(EZ, 22, "/proc/bus/usb/%s/%s", bus->dirname, dev->filename);
				//fprintf(stderr, "-> %s\n", EZ);
				
				fd = open(EZ,  O_RDWR);
				if (fd == -1) {
					perror("Open device");
					exit(1);
				}
				status = ezusb_load_ram (fd);
				if (status != 0) {
					fprintf(stderr, "Error while sending firmware\n");
					return status;
				}

				firmware_loaded=1;

				sleep(2);	// give the EZUSB some time to renumerate
				close(fd);
				printf("OK\n");
				goto find_f2a;
			}
		}
	}


	if(f2a==NULL) {
		fprintf(stderr, "error: couldn't find f2a attached to usb\n");
		return -1;
	}

	f2ahandle=usb_open(f2a);

	usb_reset(f2ahandle);

	if (f2ahandle == NULL) {
		fprintf(stderr, "f2ahandle is NULL\n");
		return -1;
	}

	result=usb_claim_interface(f2ahandle, 0x4);
	
	if(result==-1) {
		fprintf(stderr, "usb_claim_interface: %s\n", usb_strerror());
		return -1;
	}

	result=usb_claim_interface(f2ahandle, 0x83);
	if(result==-1) {
		fprintf(stderr, "usb_claim_interface: %s\n", usb_strerror());
		return -1;
	}
	return 0;
}

void usb_disconnect () {
	usb_release_interface(f2ahandle, 0);
	usb_close(f2ahandle);
}

int usb_read (char *buffer, int buffer_size) {
	int result;

	result=usb_bulk_read(f2ahandle, 0x83, buffer, buffer_size, USB_TIMEOUT);
	if(result==-1) {
		fprintf(stderr, "usb_bulk_read: %s\n", usb_strerror());
		return -1;
	}
	return result;
}

int usb_write(char *buffer, int buffer_size) {
	int result;

	result=usb_bulk_write(f2ahandle, 0x4, buffer, buffer_size, USB_TIMEOUT);
//	fprintf(stderr, "usb_write : %x, %x, %x, %x, %x => %x\n", f2ahandle, 0x4, buffer, buffer_size, USB_TIMEOUT, result);
	if(result==-1) {
		fprintf(stderr, "usb_bulk_write: %s\n", usb_strerror());
		return -1;
	}
	return result;
}


